/* [KSC4]-------------------------------------------------------------------[]
 *
 *  MAIN.C - Main module for the Power Glitch Monitor, version 0.21
 *           Written by Marco "Kiko" Carnut, 2008-Jan-04
 *
 *  Portability: Tested with avr-gcc 4.2.1 and avr-libc 1.4.6 with an ATmega8
 *  ------------ device.
 *
 * []-----------------------------------------------------------------------[]
 */

#ifndef __AVR_ATmega8__
#error Hey, this code was only tested in an ATmega8
#endif

#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>
#include <avr/sleep.h>
#include <string.h>
#include <stdlib.h>

#define F_CPU                8000000L

#include <util/delay.h>

/* --------------------------- Macro-Defined Constants -------------------- */

#define MAXTXFIFOLENGTH          120

#define TRUE                       1
#define FALSE                      0

#define FLAG_SERIAL_ECHO_OVERRIDE  1
#define FLAG_HAS_SECOND            2
#define FLAG_HAS_ANOMALY           4
#define FLAG_SERIAL_ECHO           8
#define FLAG_SERIAL_LOG_ENABLED   32
#define FLAG_CLOCK_ADJUST         64

#define MAXARGS                    3

/* ---------------------------- Function Prototypes ----------------------- */

void xmit_str(char *str);
void xmit_hex(unsigned short val, unsigned char ndigits);
void xmit_int_dp(unsigned short val, unsigned char ndigits, char point);
void dec2digits(unsigned short val, unsigned char ndigits, char point);
void xmit_clock(volatile unsigned char *c, unsigned char max, unsigned char month_name);
void xmit_str_p(prog_char *str);
void xmit(char c);
void do_xmit(char c);
void handle_serial_input(void);
void update_elapsed(unsigned char *c);
void xmit_elapsed(unsigned char *c);
void do_xmit_str_p(prog_char *str);
void update_clock(volatile unsigned char *c, unsigned char inc0);

void execute_command(void);
void cmd_date(unsigned char argc, char **argv);
void cmd_time(unsigned char argc, char **argv);
void cmd_dump(unsigned char argc, char **argv);
void set_time_date(unsigned char argc, char **argv, unsigned char base);
void cmd_log(unsigned char argc, char **argv);
void cmd_clear(unsigned char argc, char **argv);
void cmd_start(unsigned char argc, char **argv);
void cmd_finish(unsigned char argc, char **argv);
void cmd_event(unsigned char argc, char **argv);
void cmd_poke(unsigned char argc, char **argv);
void cmd_fill(unsigned char argc, char **argv);
void cmd_reboot(unsigned char argc, char **argv);
void cmd_param(unsigned char argc, char **argv);
void cmd_ver(unsigned char argc, char **argv);
void show_usage(prog_char *p);

/* ---------------------- Global Variables and Declarations --------------- */

volatile char txfifo[MAXTXFIFOLENGTH];
volatile unsigned char fifo_head;
volatile unsigned char fifo_tail;

typedef union {
    volatile unsigned short as_u16[2];
    volatile unsigned long  as_u32;
} u32_16; 

volatile u32_16 total_ticks;

volatile unsigned long t1_ticks;
volatile unsigned char semicycles;

volatile unsigned char flags;

volatile unsigned short voltage;
volatile unsigned char adc_count;

enum anomaly_t {
    ANOMALY_NONE           = 0,
    ANOMALY_POWER_OUTAGE   = 1,
    ANOMALY_VOLTAGE_SAG    = 2,
    ANOMALY_VOLTAGE_SURGE  = 3,
    ANOMALY_FREQUENCY_HIGH = 4,
    ANOMALY_FREQUENCY_LOW  = 8
} anomaly_type,prev_anomaly_type;

// The maximum values for each of those counters
unsigned char limits[] PROGMEM = { 99,59,59,23,31,12,99 };

// The starting values for each counter
unsigned char base[] PROGMEM = { 0,0,0,0,1,1,0 };

#define SUBSEC 0
#define SEC    1
#define MIN    2
#define HOUR   3
#define DAY    4
#define MONTH  5
#define YEAR   6

volatile unsigned char rtc_crystal[11];
volatile unsigned char dozens;
volatile unsigned char millissecs;
unsigned char rtc_last_failure[11];
unsigned char rtc_elapsed[4];
unsigned char cmd_set_clock_index,cmd_temp;

char digits[7];

unsigned short nominal_voltage=220;
unsigned short adc_min=480;

char month_names[] PROGMEM = "Jan\0Feb\0Mar\0Apr\0May\0Jun\0"
                             "Jul\0Aug\0Sep\0Oct\0Nov\0Dec";

unsigned char ndays_in_month[] PROGMEM = { 
    31,28,31,30,31,30,31,31,30,31,30,31
};


#define MAXCMDLINELENGTH     30

char cmdline[MAXCMDLINELENGTH];
unsigned char cmdline_cursor,cmd_in_execution;

char cmdnames[] PROGMEM = "date\0time\0reboot\0log\0ver\0";

void (*cmdvects[])(unsigned char, char **) PROGMEM = {
    cmd_date,cmd_time,cmd_reboot,cmd_log,cmd_ver
};

short nevents=-1;
unsigned char within_event;

#define enable_serial_echo()         (flags |=  FLAG_SERIAL_ECHO)
#define disable_serial_echo()        (flags &= ~FLAG_SERIAL_ECHO)
#define is_serial_echo_enabled()     (flags &   FLAG_SERIAL_ECHO)

#define enable_serial_override()     (flags |=  FLAG_SERIAL_ECHO_OVERRIDE)
#define disable_serial_override()    (flags &= ~FLAG_SERIAL_ECHO_OVERRIDE)
#define is_serial_override_enabled() (flags &   FLAG_SERIAL_ECHO_OVERRIDE)


/* -------------------------- Macro-Defined Functions --------------------- */

#define xmit_pstr(str)       xmit_str_p(PSTR(str))
#define do_xmit_pstr(str)    do_xmit_str_p(PSTR(str))
#define usage(str)           show_usage(PSTR(str))

#define xmit_int(a,b)        xmit_int_dp((a),(b),-1)

#define xmit_interrupt_enable() \
    (UCSRB |= _BV(UDRIE))
#define xmit_interrupt_disable() \
    (UCSRB &= ~_BV(UDRIE))

#define analog_comparator_bit() ((ACSR & _BV(ACO))==_BV(ACO))

/* -------------------------------- Main Program -------------------------- */

int main(void)
{
    unsigned char n;
//  DDRC |= _BV(PC5);

    /* UART initialization */
    UCSRB  =  _BV(TXEN) | _BV(RXEN) // Enable transmitter and receiver,
           |  _BV(RXCIE);           // and receiver interrupts
    UBRRL  = 51;                    // Set 9600bps for internal 8MHz RC osc

    /* Timer1 Setup -- handles waveform capture: freq & volt measurement */
    TCCR1B  = _BV(ICNC1)            // ICP noise canceler on
            | 3;                    // prescaler = 64, clk(T1) = 125,000 Hz

    OCR1A = 600;

    /* Timer2 Setup -- handles the crystal-based real time clock */
    TCCR2 = _BV(WGM21)              // Enable CTC mode
          | 4;                      // perscaler = 64, clk(T2) = 125,000Hz
    OCR2 = 124;                     // Divide by 125, OC match each 1000Hz

    TIMSK  |= _BV(TICIE1)           // Enable Input Capture interrupt
           |  _BV(OCIE1A)           // Enable Output Compare 1A interrupt
           |  _BV(OCIE2);           // Enable Timer2 Output Compare (RTC)

    /* Disable Analog Comparator, we don't use it for anything */
    ACSR |= _BV(ACD);

    /* TWI initialization */
    PORTC |= _BV(PC4) | _BV(PC5);   // Enable Pull-ups on SCL and SDA
    TWSR = 0;
    TWBR = 10;

    // Initialize RTC
    for (n=0;n<7;n++) {
        rtc_crystal[n]=pgm_read_byte(base+n);
    }
    enable_serial_echo();
    xmit_str("\r\n\r\n");
    cmd_ver(0,NULL);
    xmit_str("\r\n> ");
    for (;;) {
        /* Enable interrupts */
        sei();
        set_sleep_mode(SLEEP_MODE_IDLE);
        sleep_mode();
        asm("nop");
        if (flags & FLAG_HAS_SECOND) {
            flags &= ~FLAG_HAS_SECOND;
            if (flags & FLAG_SERIAL_LOG_ENABLED)
                enable_serial_echo();
            else
                disable_serial_echo();
            if ((anomaly_type & 3) == ANOMALY_POWER_OUTAGE) {
                xmit_pstr("* AC POWER OFFLINE *");
                xmit(' ');
            } else {
                unsigned long base=62500-520*(120-semicycles);
                base*=60000;
                unsigned short f=base/(t1_ticks>>1);
                xmit_int_dp(semicycles,4,-1); xmit(' ');
                semicycles=0;
                t1_ticks=0;

                xmit_int_dp(nominal_voltage,3,-1);
                xmit(':'); xmit(' ');
                u32_16 v;
#ifndef CALLIBRATION
                v.as_u32=(unsigned long)voltage*2978;
                xmit_int_dp(v.as_u16[1],4,1);
                xmit_pstr("V ");
#else
                xmit_int_dp(voltage,5,1);
                xmit_pstr("U  ");
                xmit_int_dp(adc_count,3,-1);
                xmit_str("  OCR1A=");
                xmit_int_dp(OCR1A,4,-1);
                xmit_str("  ");
#endif // CALLIBRATION
                xmit_int_dp(f,5,3);
                xmit_pstr("Hz ");
                voltage=0;
                adc_count=0;
            }
            xmit_clock(rtc_crystal,6,1);
            if (nevents>=0) {
                xmit(' ');
                xmit('0'+anomaly_type);
                xmit(' ');
                xmit_int_dp(nevents+1,4,-1);
                xmit_pstr(": ");
                xmit_elapsed(rtc_elapsed); xmit(' ');
                xmit_clock(rtc_last_failure,7,0);
            }
            xmit_pstr("\r\n");
        }
        if (anomaly_type != ANOMALY_NONE) {
            if (prev_anomaly_type == ANOMALY_NONE) {
                nevents++;
                memset(rtc_elapsed,0,sizeof(rtc_elapsed));
                memcpy(rtc_last_failure,(unsigned char *)rtc_crystal,sizeof(rtc_crystal));
                prev_anomaly_type=anomaly_type;
                within_event=TRUE;
            }
        } else {
            if (prev_anomaly_type != ANOMALY_NONE) {
                prev_anomaly_type=ANOMALY_NONE;
                within_event=FALSE;
            }
        }
        if (cmd_in_execution==1) execute_command();
    }

    /* We never return, but we put it here to avoid gcc warnings */
    return 0;
}

void update_clock(volatile unsigned char *c, unsigned char inc0)
{
    if (flags & FLAG_CLOCK_ADJUST) return;
    unsigned char i;
    unsigned char inc=inc0;
    for (i=0;i<7;i++) {
        unsigned char limit=pgm_read_byte(limits+i);
        if (i==DAY) {
            // Handle varying month lengths and leap years
            // This will work up to 2099
            limit  = pgm_read_byte(c[MONTH]+ndays_in_month-1);
            limit += (c[MONTH]==2) && !(c[YEAR]&3);
        }
        c[i]+=inc;
        if (c[i]<=limit) break;
        c[i]=pgm_read_byte(base+i);
        inc=1;
    }
    // Increment linear counter as well -- simplifies
    // calculating time differences
    unsigned long *p=(unsigned long *)(c+7);
    (*p)+=inc0;
}

void update_elapsed(unsigned char *c)
{
    unsigned char i;
    for (i=0;i<4;i++) {
        unsigned char limit=pgm_read_byte(limits+i);
        if (i==HOUR) limit=255;
        if (++c[i]<=limit) break;
        c[i]=0;
    }
}

void xmit_clock(volatile unsigned char *c, unsigned char max, unsigned char month_name)
{
    prog_char *sep =  PSTR(".:: --");
    unsigned char n;
    for (n=0;n<max;n++) {
        unsigned char s=pgm_read_byte(sep+6-n);
        if (s) xmit(s);
        if (n==1 && month_name) {
            xmit_str_p(month_names+((c[5]-1)<<2));
        } else {
            xmit_int(c[6-n],2);
        }
    }
}

void xmit_elapsed(unsigned char *c)
{
    xmit_int_dp(c[HOUR],3,-1);
    xmit('h');
    xmit_int_dp(c[MIN],2,-1);
    xmit('m');
    xmit_int_dp(c[SEC],2,-1);
    xmit('s');
    xmit_int_dp(c[SUBSEC],2,-1);
}

SIGNAL(SIG_OUTPUT_COMPARE1A)
{
    ADMUX  = 0;                     // External reference, channel 0
    ADCSRA = _BV(ADEN) | _BV(ADIE)  // Enable ADC, enable ADC interrupt
           | _BV(ADSC)              // Start ADC
           | 6;                     // prescaler = 64, clk(ADC) = 125KHz
}

SIGNAL(SIG_INPUT_CAPTURE1)
{
    TCNT1=0;
    t1_ticks+=ICR1;
    semicycles++;
}

SIGNAL(SIG_ADC)
{
    unsigned short adc=ADC;
    voltage+=adc>>1;
    ADCSRA = 0; // Disable ADC entirely
    adc_count++;
    enum anomaly_t anomaly=ANOMALY_NONE;
    if (adc<adc_min) {
        anomaly=ANOMALY_POWER_OUTAGE;
        TIMSK &= ~_BV(TICIE1);
    }
    anomaly_type = (anomaly_type & 0xFC) | anomaly;
    if (!(TIMSK & _BV(TICIE1))) {
        TCNT1=0; semicycles=0;
        if (anomaly==ANOMALY_NONE) { 
            TIMSK |= _BV(TICIE1);
        }
    }
#ifdef CALLIBRATION
    PORTC &= ~_BV(PC0);
    DDRC  |=  _BV(PC0);
    _delay_us(10);
    DDRC  &= ~_BV(PC0);
#endif // CALLIBRATION
}

SIGNAL(SIG_OUTPUT_COMPARE2)
{
    if (++millissecs==10) {
        millissecs=0;
        update_clock(rtc_crystal,1);
        if (!rtc_crystal[SUBSEC]) flags |= FLAG_HAS_SECOND;
        if (within_event) update_elapsed(rtc_elapsed);
    }
}

void dec2digits(unsigned short val, unsigned char ndigits, char point)
{
    unsigned char n;
    if (point>=0) {
        ndigits++;
        point=ndigits-point;
    }
    for (n=ndigits;n>0;n--) {
        if (n==point) {
            digits[n-1]='.';
        } else {
            digits[n-1]=val%10+'0';
            val/=10;
        }
    }
    digits[ndigits]=0;
}

void xmit_int_dp(unsigned short val, unsigned char ndigits, char point)
{
    dec2digits(val,ndigits,point);
    xmit_str(digits);
}

void xmit_hex(unsigned short val, unsigned char ndigits)
{
    char d[5];
    unsigned char n;
    for (n=ndigits;n>0;n--) {
        char c=(val&0xF)+'0';
        if (c>'9') c+=7;
        d[n-1]=c;
        val>>=4;
    }
    d[ndigits]=0;
    xmit_str(d);
}

void xmit_hex_str(unsigned char *p, int length)
{
    while (length--) xmit_hex(*p++,2);
}

SIGNAL(SIG_UART_DATA)
{
    if (fifo_tail==fifo_head) {
        xmit_interrupt_disable();
    } else {
        UDR = txfifo[fifo_tail++];
        if (fifo_tail==MAXTXFIFOLENGTH) fifo_tail=0;
    }
}

SIGNAL(SIG_UART_RECV)
{
    unsigned char c=UDR;
    if (flags & FLAG_SERIAL_LOG_ENABLED) {
        flags &= ~FLAG_SERIAL_LOG_ENABLED;
        do_xmit_pstr("\r\n> ");
        return;
    }
    if (!cmd_in_execution) {
        if (c==13) {            // Enter
            cmd_in_execution=1; // Dispatcher
            do_xmit_pstr("\r\n");
        } else if (c==3) {
            do_xmit_pstr("\r\n> ");
            cmdline_cursor=0;
        } else if (c==8) {
            if (cmdline_cursor) {
                do_xmit_pstr("\b \b");
                cmdline_cursor--;
            }
        } else if (c>=' ' && c<=127) {
            if (cmdline_cursor<MAXCMDLINELENGTH) {
                cmdline[cmdline_cursor++]=c;
                do_xmit(c);
            }
        }
    }
}

void xmit_str(char *str)
{
    for (;*str;str++) xmit(*str);
}

void xmit_str_p(prog_char *str)
{
    for (;;) {
        unsigned c=pgm_read_byte(str++);
        if (!c) break;
        xmit(c);
    }
}

void do_xmit_str_p(prog_char *str)
{
    enable_serial_override();
    xmit_str_p(str);
    disable_serial_override();
}

void do_xmit(char c)
{
    unsigned char next_head=fifo_head+1;
    if (next_head==MAXTXFIFOLENGTH) next_head=0;
    /* If there is still room in the FIFO, add character by character */
    if (next_head==fifo_tail) return;
    txfifo[fifo_head]=c;
    fifo_head=next_head;
    xmit_interrupt_enable();
}

void xmit(char c)
{
    if (is_serial_echo_enabled() || is_serial_override_enabled())
        do_xmit(c);
}

void execute_command(void)
{
    cmd_in_execution=2;
    unsigned char vect=0;
    prog_char *p=cmdnames;
    prog_char *line=cmdline;
    cmdline[cmdline_cursor]=0;
    // Ignore leading spaces
    while (*line==' ') line++,cmdline_cursor--;
    for (;cmdline_cursor;) {
        if (!pgm_read_byte(p)) {
            do_xmit_pstr("Unrecognized command");
            break;
        }
        if (!strncmp_P(line,p,strlen_P(p))) {
            unsigned char argc=0;
            char *argv[MAXARGS];
            while (*line && *line!=' ') line++;
            for (;*line;line++) {
                if (*line==' ') {
                    *line=0;
                } else {
                    if (!line[-1] && argc<MAXARGS) argv[argc++]=line;
                }
            }
            // Jumps to the function that handles that command, passing
            // the number of arguments and the separated words in the
            // command line. This code is a bit unreadable, but it simply
            // fetches the function address from the cmdvects array (indexed
            // by vect), converts the word to a function pointer and calls
            // the address pointed to by this pointer with argc and argv as
            // addresses.
            ((void (*)(int,char **))pgm_read_word(cmdvects+vect))(argc,argv);
            break;
        }
        while (pgm_read_byte(p)) p++;
        p++; vect++;
    }
    do_xmit_pstr("\r\n> ");
    cmd_in_execution=0; cmdline_cursor=0;
}

void cmd_date(unsigned char argc, char **argv)
{
    if (argc!=3) {
        usage("date yy mm dd");
        return;
    }
    set_time_date(3,argv,YEAR);
}

void cmd_time(unsigned char argc, char **argv)
{
    if (argc!=3 && argc!=2) {
        usage("time hh mm [ss]");
        return;
    }
    set_time_date(argc,argv,HOUR);
}

void set_time_date(unsigned char argc, char **argv, unsigned char base)
{
    unsigned char i;
    flags |= FLAG_CLOCK_ADJUST;
    for (i=0;i<argc;i++) {
        rtc_crystal[base-i]=atoi(argv[i]);
    }
    if (base==HOUR) {
        rtc_crystal[SUBSEC]=0;
        if (argc==2) rtc_crystal[SEC]=0;
    }
    flags &= ~FLAG_CLOCK_ADJUST;
}

void cmd_log(unsigned char argc, char **argv)
{
    flags |= FLAG_SERIAL_LOG_ENABLED;
}

void cmd_clear(unsigned char argc, char **argv)
{
    cli();
    within_event=FALSE;
    semicycles=0;
    sei();
}

void cmd_reboot(unsigned char argc, char **argv)
{
    cli();
    WDTCR |= _BV(WDE);
    for (;;);
}

void cmd_ver(unsigned char argc, char **argv)
{
    do_xmit_pstr("Power Minder v0.22 - (C) 2008 by Kiko built on "
     __DATE__ " " __TIME__ "\r\n");
}

void show_usage(prog_char *p)
{
    do_xmit_pstr("Usage: ");
    do_xmit_str_p(p);
    do_xmit_pstr("\r\n");
}
